package DBIx::Class::InflateColumn::TimeMomentX_BOKUTIN;

use strict;
use warnings;
use base qw/DBIx::Class/;
use DBIx::Class::Carp;
use Time::MomentX::BOKUTIN;
use namespace::clean;

# https://fastapi.metacpan.org/source/RIBASUSHI/DBIx-Class-0.082842/lib/DBIx/Class/InflateColumn/DateTime.pm
# https://metacpan.org/pod/DateTime::Format::Pg

sub register_column {
  my ($self, $column, $info, @rest) = @_;

  $self->next::method($column, $info, @rest);

  my $data_type = $info->{data_type} || '';
  return unless $data_type =~ /^(date|datetime|timestamp|timestamp with time zone)$/; # time は扱わない

  my $precision = $info->{size} || ($data_type =~ /time/ ? 6 : undef);

  $self->inflate_column(
    $column =>
      {
        inflate => sub {
          my ($value, $obj) = @_;

          local $@;
          my $tm = eval { tmx($value) };
          $self->throw_exception("Error while inflating '$value' for $column on ${self}. error:$@") if $@;
          $tm;
        },
        deflate => sub {
          # my ($value, $obj) = @_;

          my $tm = $_[0]->isa('Time::Moment') ? $_[0] : tmx( from_object => $_[0] );

          # ->with_system_offset_same_instant する理由
          #
          # offset は静かに棄てられてしまうので日時を変換しておきます。
          # with_system_offset_same_instant は tzset(2) に従うだけなので show timezone と一致しているかは保証しません。
          #
          # 変換のコストを避けたい場合は、事前に文字列にして、ここを通らないようにしてください。
          #
          # 既に system_offset になっているか確認して呼ばない分岐を入れる必要が無いくらい速いはずなので、ただ呼ぶだけにする

          if ($data_type eq 'date') {
              $tm->with_system_offset_same_instant->strftime('%F');
          }
          elsif ($data_type eq 'timestamp') {
              _pg_like_round($tm->with_system_offset_same_instant,$precision)->strftime("%F %T%f");
          }
          elsif ($data_type eq 'timestamp with time zone') {
              _pg_like_round($tm->with_system_offset_same_instant,$precision)->strftime("%F %T%f%Z") =~ s/:00$//r;
          }
          else {
              # Pgにおいては、できれば避けたい。
              # pg側での受け入れとしての解釈には問題なくとも
              # selectした場合の文字列と表現が異なると（意味は同じでも）is_changed などが 1 になってしまう。
              $tm->to_string;
          }
        },
      }
  );
}

sub _pg_like_round {
    my ($tm, $precision) = @_;

    if ( $tm->precision > $precision ) {
        # 文字列から取り出しても、数値で変換しても、速度には大差ない
        # my $num = substr $tm->nanosecond, -(9-$precision), 1;
        my $num = int $tm->nanosecond % (10**(9-$precision)) / (10**(9-$precision-1)); # 桁を取り出す
        if ( $num >= 5 ) {
            # nanosecond
            # 1,000,000,000 1sec
            my $ns = 10**(9-$precision);
            $tm = $tm->plus_nanoseconds($ns);
        }
        $tm = $tm->with_precision($precision);
    }

    $tm;
}

1;

__END__

DBIx::Class::InflateColumn::DateTime と DateTime::Format::Pg はすばらしい

    DBIx::Class::InflateColumn::DateTime はとても柔軟に出来ていますが、ここでは色々省きます。

postgres の動き

    without timezone
        取得時
            offset は付きません。
                2018-01-01 00:00:00
        保存時
            Z や +09:00 といった offset は静かに棄てられます！
            show timezone に変換してくれたりはしないようです。
                2018-01-01T00:00:00+09:00 -> 2018-01-01 00:00:00
                2018-01-01T00:00:00Z      -> 2018-01-01 00:00:00

            https://www.postgresql.jp/document/9.5/html/datatype-datetime.html
            timestamp without time zoneと決定済みのリテラルでは、PostgreSQLは警告なく時間帯情報をすべて無視します。 つまり、結果の値は明示された入力値の日付/時刻フィールドから持ち込まれますが、時間帯の調整はなされません。

    with timezone
        取得時
            set timezone の offset が付きます。
                2018-01-01 00:00:00+09 （末尾の :00 は省略されます）
        保存時
            offset を解釈して、内部的には UTC で保存されるようです。

この InflateColumn では

    without time zone でも show timezone を確認しません。

    ただ tmx に委ねます。

備考

    show timezone と tzset(2) が一致しないケースは、今のところサポートしていません。
    その必要がある場合には show timezone と tzset(2) を一致させた上 without -> with などを検討してください。
    https://stackoverflow.com/questions/9772825/postgresql-alter-type-timestamp-without-time-zone-with-time-zone
